#version 300 es

//-----------------------------------------------------------------------
// Copyright (c) 2018 Snap Inc.
//-----------------------------------------------------------------------

// We need to fix assembleShaderSource method. Becuase it paste require.glsl content at start of file,
// so it header occur before #version. But #version should be first line of shader code.
#ifdef HACK
#include <required.glsl>
#endif

// We implemet this extensions support in required.glsl, but we can not include it because we use version 300 in header for ES3.0
// And if we include required.glsl then we get error messages, because shader builder paste required.glsl code before version 300 es header
#if defined(GL_ARB_draw_instanced)
#extension GL_ARB_draw_instanced : require
#endif // ARB_draw_instanced

#if defined(GL_EXT_draw_instanced)
#extension GL_EXT_draw_instanced : require
#endif // GL_EXT_draw_instanced

#define SC_USE_USER_DEFINED_VS_MAIN

#if __VERSION__ >= 300
precision highp int;
precision highp float;
#else
#define texture texture2D
#endif

uniform mat4 sc_ViewMatrix;
uniform mat4 sc_ProjectionMatrix;

uniform sampler2D hairTexture;
uniform sampler2D litLookupTexture;
uniform sampler2D strandDataMapTexture;
uniform vec4 strandDataMapTextureSize;

uniform vec4 hairColor;
uniform vec4 diffuseLightColor;
uniform vec4 specularLightColor;
uniform float alphaThreshold;

uniform float strandWidth;
uniform float strandBunchAberration;
uniform float singleStrandInterpolationRadius;
uniform float singleStrandInterpolationInstanceCount; // instance count per one strand on single strand interpolation
uniform float multiStrandInterpolationRadius;
uniform float multiStrandInterpolationInstanceCount; // instance count per one strand on multi strand interpolation
uniform float isDebugMode;

#ifdef VERTEX_SHADER

#if __VERSION__ < 300
attribute vec4 position;
varying vec4 debugColor; // For debug mode colors draw
varying vec4 hairTextureUV;
#else
in vec4 position;
out vec4 debugColor; // For debug mode colors draw
out vec4 hairTextureUV;
#endif // __VERSION__

// platform dependent instance id getter
int getInstanceID() {
#if __VERSION__ >= 300
    return gl_InstanceID;
#elif defined(GL_ARB_draw_instanced)
    return gl_InstanceIDARB;
#elif defined(GL_EXT_draw_instanced)
    return gl_InstanceIDEXT;
#else
#error Current platform not supported instancing
#endif
}

// pseudo random values in range [0.0; 1.0]
float rand(vec2 co) {
    return fract(sin(dot(co.xy, vec2(12.9898, 78.233))) * 43758.5453);
}

void main(void) {
    float instanceId = float(getInstanceID());
    float strandInstanceCount = 1.0 + singleStrandInterpolationInstanceCount + multiStrandInterpolationInstanceCount;
    float instanceCount = strandDataMapTextureSize.y * strandInstanceCount; // instance count of all draw call instances
    
    const float epsilon = 0.0001;
    float basicStrandIndex = floor(instanceId / strandInstanceCount + epsilon); // basic strand index
    float basicStrandInstanceId = basicStrandIndex * strandInstanceCount;
    float relativeInstanceId = instanceId - basicStrandInstanceId;
    bool isSingleInterpolationStrand = relativeInstanceId > 0.0 && relativeInstanceId <= singleStrandInterpolationInstanceCount;
    bool isMultiInterpolationStrand = relativeInstanceId > singleStrandInterpolationInstanceCount && relativeInstanceId < strandInstanceCount;
    bool isBasicStrand = !isSingleInterpolationStrand && !isMultiInterpolationStrand;
    
    vec2 uvAbsoluteMax = vec2(strandDataMapTextureSize.x - 1.0, strandDataMapTextureSize.y - 1.0);
    vec2 uvOffsetX = vec2(1.0 / uvAbsoluteMax.x, 0.0); // Offset to move through strand points
    
    // interpolation random terms
    vec2 randomTerms = vec2(instanceId / instanceCount, (instanceCount - instanceId) / instanceCount);
    
    // uv: x - strand point at strandDataMapTexture, y - strand at strandDataMapTexture
    vec2 uv = vec2(position.y / uvAbsoluteMax.x, basicStrandIndex / uvAbsoluteMax.y);
    vec3 strandPointOffset = vec3(0.0);
    
    // Get strand point world position
    vec3 pointWorldPosition = texture(strandDataMapTexture, uv).xyz;
    
    // Calculate strand point position offset for multi strand interpolation
    if (isMultiInterpolationStrand) {
        // Neighbour indices stored in the first two strand point alpha components
        vec2 neighbourIndices = vec2(-1.0);
        neighbourIndices.x = texture(strandDataMapTexture, vec2(0.0, uv.y)).w;
        neighbourIndices.y = texture(strandDataMapTexture, vec2(uvOffsetX.x, uv.y)).w;
        
        if (neighbourIndices.x < 0.0 && neighbourIndices.y < 0.0) {
            isSingleInterpolationStrand = true;
            isMultiInterpolationStrand = false;
        } else {
            vec3 pointWorldPositionInterpolated = pointWorldPosition;
            
            if (neighbourIndices.x >= 0.0) {
                vec2 uv1 = vec2(uv.x, neighbourIndices.x / uvAbsoluteMax.y);
                vec3 pointWorldPosition = texture(strandDataMapTexture, uv1).xyz;
                float lerpTerm = rand(randomTerms);
                pointWorldPositionInterpolated = mix(pointWorldPositionInterpolated, pointWorldPosition, lerpTerm);
            }
            
            if (neighbourIndices.y >= 0.0) {
                vec2 uv2 = vec2(uv.x, neighbourIndices.y / uvAbsoluteMax.y);
                vec3 pointWorldPosition = texture(strandDataMapTexture, uv2).xyz;
                float lerpTerm = rand(randomTerms.xx);
                pointWorldPositionInterpolated = mix(pointWorldPositionInterpolated, pointWorldPosition, lerpTerm);
            }
            
            strandPointOffset = pointWorldPositionInterpolated - pointWorldPosition;
        }
    }
    
    // Calculate strand point position offset for single strand interpolation
    if (isSingleInterpolationStrand) {
        vec3 randomOffset = vec3(0.0);
        randomOffset.x = rand(randomTerms.xx) * 2.0 - 1.0;
        randomOffset.y = rand(vec2(randomOffset.x, randomTerms.x)) * 2.0 - 1.0;
        randomOffset.z = rand(vec2(randomOffset.x, randomOffset.y)) * 2.0 - 1.0;
        
        float strandBunchAberrationTerm = uv.x * strandBunchAberration; // strand point position aberration
        strandPointOffset = randomOffset * (singleStrandInterpolationRadius + strandBunchAberrationTerm);
    } else if (!isMultiInterpolationStrand && isDebugMode > 0.0) {
        // For Debug (float precision problems)
        strandPointOffset.x += (relativeInstanceId != 0.0 ? 3.0 : 0.0);
    }
    
    // Get strand point view position
    pointWorldPosition += strandPointOffset;
    vec4 pointViewPosition = sc_ViewMatrix * vec4(pointWorldPosition.x, pointWorldPosition.y, pointWorldPosition.z, 1.0);
    
    // Calculate triangle strip vertex positions by strand width and offset position
    vec2 smoothedTangent = vec2(0.0);
    
    if (position.y < uvAbsoluteMax.x) { // Check for strand point is not last
        vec3 pointWorldPositionNext = texture(strandDataMapTexture, uv + uvOffsetX).xyz + strandPointOffset;
        vec4 pointViewPositionNext = sc_ViewMatrix * vec4(pointWorldPositionNext.x, pointWorldPositionNext.y, pointWorldPositionNext.z, 1.0);
        smoothedTangent = normalize(pointViewPositionNext.xy - pointViewPosition.xy);
    }
    
    if (position.y > 0.0) { // Check for strand point is not first
        vec3 pointWorldPositionPrevious = texture(strandDataMapTexture, uv - uvOffsetX).xyz + strandPointOffset;
        vec4 pointViewPositionPrevious = sc_ViewMatrix * vec4(pointWorldPositionPrevious.x, pointWorldPositionPrevious.y, pointWorldPositionPrevious.z, 1.0);
        smoothedTangent += normalize(pointViewPosition.xy - pointViewPositionPrevious.xy);
    }
    
    if (position.y > 0.0 && position.y < uvAbsoluteMax.x) {
        smoothedTangent *= 0.5;
    }
    vec2 smoothedBinormal = vec2(smoothedTangent.y, -smoothedTangent.x);
    
    pointViewPosition.xy += smoothedBinormal * position.x * strandWidth;
    hairTextureUV.xy = vec2(-position.x * 0.5 + 0.5, 1.0 - uv.x);
    hairTextureUV.zw = abs(smoothedBinormal.xy);
    
    // For debug
    debugColor = vec4(1.0);
    if (isDebugMode > 0.0) {
        if (isBasicStrand) {
            debugColor.rgb = vec3(0.0, 1.0, 0.0);
        } else if (isSingleInterpolationStrand) {
            debugColor.rgb = vec3(0.0, 0.0, 1.0);
        } if (isMultiInterpolationStrand) {
            debugColor.rgb = vec3(1.0, 0.0, 0.0);
        }
    }
    // ---------
    
    gl_Position = sc_ProjectionMatrix * pointViewPosition;
}

#endif // VERTEX_SHADER


#ifdef FRAGMENT_SHADER

#if __VERSION__ < 300
varying vec4 debugColor; // For debug
varying vec4 hairTextureUV;
#else
in vec4 debugColor; // For debug
in vec4 hairTextureUV;
out vec4 fragColor;
#endif // __VERSION__

void main(void) {
    vec4 hairTextureSample = texture(hairTexture, hairTextureUV.xy);
    if (hairTextureSample.a < alphaThreshold) {
        discard;
    }
    
    if (isDebugMode > 0.0) {
        hairTextureSample.rgb = debugColor.rgb;
    } else {
        vec4 litLookupTextureSample = texture(litLookupTexture, hairTextureUV.zw);
        hairTextureSample.rgb = mix(hairTextureSample.rgb, hairTextureSample.rgb * hairColor.rgb, hairColor.a);
        hairTextureSample.rgb = mix(hairTextureSample.rgb, hairTextureSample.rgb * diffuseLightColor.rgb * litLookupTextureSample.r, diffuseLightColor.a);
        hairTextureSample.rgb = mix(hairTextureSample.rgb, hairTextureSample.rgb + specularLightColor.rgb * litLookupTextureSample.g, specularLightColor.a);
    }
    
#if __VERSION__ < 300
    gl_FragColor = hairTextureSample;
#else
    fragColor = hairTextureSample;
#endif
}

#endif //FRAGMENT SHADER
