//-----------------------------------------------------------------------
// Copyright (c) 2019 Snap Inc.
//-----------------------------------------------------------------------

#define SC_USE_USER_DEFINED_VS_MAIN

uniform mat4 sc_ProjectionMatrix;
uniform mat4 sc_ModelViewMatrix;

uniform sampler2D hairTexture;
uniform sampler2D litLookupTexture;

uniform vec4 hairColor;
uniform vec4 diffuseLightColor;
uniform vec4 specularLightColor;
uniform float alphaThreshold;

uniform float strandWidth;
uniform float isDebugMode;

varying vec4 hairTextureUV;
varying vec4 debugColor; // For debug mode colors draw

#ifdef VERTEX_SHADER

attribute vec3 position;
attribute vec3 positionNext;
attribute vec3 positionPrevious;
attribute vec3 strandProperties;
attribute vec2 texture0;

// strandProperties:
// x - binormal sign (-1 || 1)
// y - strand point index mark:
//     0 - point placed between first and last points
//     1 - first point
//     2 - last point
// z - strand type:
//     0 - basic strand
//     1 - single strand interpolation
//     2 - multi strand interpolation

void main(void) {
    vec4 pointViewPosition = sc_ModelViewMatrix * vec4(position.x, position.y, position.z, 1.0);
    vec4 pointViewPositionNext = sc_ModelViewMatrix * vec4(positionNext.x, positionNext.y, positionNext.z, 1.0);
    vec4 pointViewPositionPrevious = sc_ModelViewMatrix * vec4(positionPrevious.x, positionPrevious.y, positionPrevious.z, 1.0);
    
    vec2 smoothedTangent = vec2(0.0);
    float strandPointIndexMark = strandProperties.y;
    
    if (strandPointIndexMark != 2.0) { // if strand point is not last
        smoothedTangent += normalize(pointViewPositionNext.xy - pointViewPosition.xy);
    }
    if (strandPointIndexMark != 1.0) { // if strand point is not first
        smoothedTangent += normalize(pointViewPosition.xy - pointViewPositionPrevious.xy);
    }
    if (strandPointIndexMark == 0.0) { // if strand point is not first and not last
        smoothedTangent *= 0.5;
    }
    
    float binormalSign = strandProperties.x;
    vec2 smoothedBinormal = vec2(smoothedTangent.y, -smoothedTangent.x);
    
    pointViewPosition.xy += smoothedBinormal * binormalSign * strandWidth;
    hairTextureUV.xy = texture0;
    hairTextureUV.zw = abs(smoothedBinormal.xy);
    
    // For debug mode colors draw
    float strandType = strandProperties.z;
    debugColor = vec4(1.0);
    if (isDebugMode > 0.0) {
        if (strandType == 0.0) {
            debugColor.rgb = vec3(0.0, 1.0, 0.0);
        } else if (strandType == 1.0) {
            debugColor.rgb = vec3(0.0, 0.0, 1.0);
        } if (strandType == 2.0) {
            debugColor.rgb = vec3(1.0, 0.0, 0.0);
        }
    }
    // ---------
    
    gl_Position = sc_ProjectionMatrix * pointViewPosition;
}

#endif // VERTEX_SHADER


#ifdef FRAGMENT_SHADER

void main(void) {
    vec4 hairTextureSample = texture2D(hairTexture, hairTextureUV.xy);
    if (hairTextureSample.a < alphaThreshold) {
        discard;
    }
    
    if (isDebugMode > 0.0) {
        hairTextureSample.rgb = debugColor.rgb;
    } else {
        vec4 litLookupTextureSample = texture2D(litLookupTexture, hairTextureUV.zw);
        hairTextureSample.rgb = mix(hairTextureSample.rgb, hairTextureSample.rgb * hairColor.rgb, hairColor.a);
        hairTextureSample.rgb = mix(hairTextureSample.rgb, hairTextureSample.rgb * diffuseLightColor.rgb * litLookupTextureSample.r, diffuseLightColor.a);
        hairTextureSample.rgb = mix(hairTextureSample.rgb, hairTextureSample.rgb + specularLightColor.rgb * litLookupTextureSample.g, specularLightColor.a);
    }
    
    gl_FragColor = hairTextureSample;
}

#endif //FRAGMENT SHADER

